/*
  MQTT Bluetooth (BLE) Anwesenheitserkennung. Raum präsenzmelder für Hausautomatisierung. Personen präsenz im Zimmer durch:
  iBeacon, Mi Band fitness tracker, smartwatch aufspüren. Publish ist MQTT_Name, Daten MAC und RSSI in JSON Format alle 10s
  {
  "28:A1:83:E1:58:8E":-60,
  "ea:5f:8d:69:be:56":-80
  }

     BLEAdvertisedDevice();
 
     BLEAddress  getAddress();
     uint16_t    getApperance();
     std::string getManufacturerData();
     std::string getName();
     int         getRSSI();
     BLEScan*    getScan();
     BLEUUID     getServiceUUID();
     int8_t      getTXPower();
 
     bool        haveAppearance();
     bool        haveManufacturerData();
     bool        haveName();
     bool        haveRSSI();
     bool        haveServiceUUID();
     bool        haveTXPower();

*/


#include <WiFi.h>
#include <WiFiClient.h>
#include <PubSubClient.h>
#include <WebServer.h>
#include <Update.h>
#include <BLEDevice.h>
#include <BLEScan.h>
#include <BLEAddress.h>

String pos = "bett";

// MQTT Zugangsdaten
#define MQTT_server "192.168.192.26"
#define MQTT_port 1883
#define MQTT_User ""
#define MQTT_Pass ""
#define QoS 0
String MQTT_Name = "bleScan/";    // gleichzeitig publish Name zzgl. mac-Adresse

const char* ssid = "Unifi"; // WLAN Zugangsdaten Eintragen
const char* password = "deinWlanPasswort";

int scanTime = 7; //sek. iBeacon Suchzeit und MQTT Berichtinterval

BLEScan* pBLEScan ;

// --------------- MQTT Variablen ---------------------------------------
WiFiClient wclient;
PubSubClient mqtt(wclient);

WebServer server(80);

/*
 * Login page
 */

const char* loginIndex = 
 "<form name='loginForm'>"
    "<table width='20%' bgcolor='A09F9F' align='center'>"
        "<tr>"
            "<td colspan=2>"
                "<center><font size=4><b>ESP32 Login Page:</b></font></center>"
                "<br>"
            "</td>"
            "<br>"
            "<br>"
        "</tr>"
        "<td>Username:</td>"
        "<td><input type='text' size=25 name='userid'><br></td>"
        "</tr>"
        "<br>"
        "<br>"
        "<tr>"
            "<td>Password:</td>"
            "<td><input type='Password' size=25 name='pwd'><br></td>"
            "<br>"
            "<br>"
        "</tr>"
        "<tr>"
            "<td><input type='submit' onclick='check(this.form)' value='Login'></td>"
        "</tr>"
    "</table>"
"</form>"
"<script>"
    "function check(form)"
    "{"
    "if(form.userid.value=='admin' && form.pwd.value=='admin')"
    "{"
    "window.open('/serverIndex')"
    "}"
    "else"
    "{"
    " alert('Error Password or Username')/*displays error message*/"
    "}"
    "}"
"</script>";
 
/*
 * Server Index Page
 */
 
const char* serverIndex = 
"<script src='https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js'></script>"
"<form method='POST' action='#' enctype='multipart/form-data' id='upload_form'>"
   "<input type='file' name='update'>"
        "<input type='submit' value='Update'>"
    "</form>"
 "<div id='prg'>progress: 0%</div>"
 "<script>"
  "$('form').submit(function(e){"
  "e.preventDefault();"
  "var form = $('#upload_form')[0];"
  "var data = new FormData(form);"
  " $.ajax({"
  "url: '/update',"
  "type: 'POST',"
  "data: data,"
  "contentType: false,"
  "processData:false,"
  "xhr: function() {"
  "var xhr = new window.XMLHttpRequest();"
  "xhr.upload.addEventListener('progress', function(evt) {"
  "if (evt.lengthComputable) {"
  "var per = evt.loaded / evt.total;"
  "$('#prg').html('progress: ' + Math.round(per*100) + '%');"
  "}"
  "}, false);"
  "return xhr;"
  "},"
  "success:function(d, s) {"
  "console.log('success!')" 
 "},"
 "error: function (a, b, c) {"
 "}"
 "});"
 "});"
 "</script>";

/*
 * setup function
 */
 
void setup()
{
  int i;
  Serial.begin(115200);
  Serial.println("");
  Serial.println("Start");

  // --------------------- WiFi ---------------------------------------------
  Serial.println("Warte auf Verbindung");
  WiFi.mode(WIFI_STA);   ;
  WiFi.begin(ssid, password);
  Serial.println("");
  int timout = 0;
  while (WiFi.status() != WL_CONNECTED)
  {
    delay(500);
    Serial.print("O");
    timout++;
    if  (timout > 20) // Wenn Anmeldung nicht möglich
    {
      Serial.println("");
      Serial.println("Wlan verbindung fehlt");
      ESP.restart(); // ESP32 neu starten
    }
  }
  Serial.println("");
  Serial.print("IP Addresse: ");
  Serial.println(WiFi.localIP());

  //String mac = WiFi.macAddress();
  String ip = WiFi.localIP().toString();
  ip.replace(".", "_");
  MQTT_Name = MQTT_Name + pos + "_" + ip;

  // Starte Webserver
    /*return index page which is stored in serverIndex */
  server.on("/", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", loginIndex);
  });
  server.on("/serverIndex", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", serverIndex);
  });
  /*handling uploading firmware file */
  server.on("/update", HTTP_POST, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/plain", (Update.hasError()) ? "FAIL" : "OK");
    ESP.restart();
  }, []() {
    HTTPUpload& upload = server.upload();
    if (upload.status == UPLOAD_FILE_START) {
      Serial.printf("Update: %s\n", upload.filename.c_str());
      if (!Update.begin(UPDATE_SIZE_UNKNOWN)) { //start with max available size
        Update.printError(Serial);
      }
    } else if (upload.status == UPLOAD_FILE_WRITE) {
      /* flashing firmware to ESP*/
      if (Update.write(upload.buf, upload.currentSize) != upload.currentSize) {
        Update.printError(Serial);
      }
    } else if (upload.status == UPLOAD_FILE_END) {
      if (Update.end(true)) { //true to set the size to the current progress
        Serial.printf("Update Success: %u\nRebooting...\n", upload.totalSize);
      } else {
        Update.printError(Serial);
      }
    }
  });
  server.begin();

  // ---------------------- MQTT ---------------------------------------------------
  mqtt.set_server(MQTT_server, MQTT_port); // MQTT Port einstellen
  // if (mqtt.connect(MQTT::Connect(MQTT_Name).set_auth(MQTT_User, MQTT_Pass)))
  if (mqtt.connect(MQTT_Name))
  {
    Serial.println("Mit MQTT verbunden");
    mqtt.publish(MQTT::Publish("hallo", MQTT_Name).set_qos(QoS)); // im Topic hallo Präsenzmelder Name schreiben
  }

  Serial.println("Starte BLE Scanner");
  BLEDevice::init("");
  pBLEScan = BLEDevice::getScan();
  pBLEScan->setActiveScan(true);   // Aktives Bluetooth Scannen
  pBLEScan->setInterval(100);
  pBLEScan->setWindow(99);
}


void loop()
{
  server.handleClient();
  
  int j;
  String NachrichtJSON;
  BLEScanResults foundDevices = pBLEScan->start(scanTime); // BLE Scanen
  
  NachrichtJSON = "[";     // JSON Format vorbereiten

  // Alle gefundene MACs durchgehen
  for (j = 0; j < foundDevices.getCount(); j++) // Alle Gefundene Bluetooth Device durchgehen
  {
    String manu = "";
    int len = sizeof(foundDevices.getDevice(j).getManufacturerData());
    for (byte i = 0; i < len; i++)
      manu += String(foundDevices.getDevice(j).getManufacturerData()[i], HEX);

    NachrichtJSON = NachrichtJSON + "{\r\n";
    NachrichtJSON = NachrichtJSON + "\"manuData\": \"" + manu + "\",\r\n";
    NachrichtJSON = NachrichtJSON + "\"mac\": \"" + foundDevices.getDevice(j).getAddress().toString().c_str() + "\",\r\n";
    NachrichtJSON = NachrichtJSON + "\"rssi\": " + String(foundDevices.getDevice(j).getRSSI()) + "\r\n";
    NachrichtJSON = NachrichtJSON + "}";
    
    if (j != foundDevices.getCount() - 1)  NachrichtJSON = NachrichtJSON + ", ";
  }
  NachrichtJSON = NachrichtJSON + "]";
  pBLEScan->clearResults(); // Speicher freigeben
  mqtt.publish(MQTT::Publish(MQTT_Name, NachrichtJSON).set_qos(QoS)); // JSON per MQTT senden
  Serial.println(NachrichtJSON);
  delay(100);
  if (WiFi.status() != WL_CONNECTED)  ESP.restart(); // Wenn Wlan fehlt neustarten
  if (mqtt.connected())
  {
    mqtt.loop();
  }
  else
  {
    Serial.println("MQTT Verbindung fehlt"); // Wenn MQTT Verbindung fehlt dann neu verbinden
    // mqtt.connect(MQTT::Connect(MQTT_Name).set_auth(MQTT_User, MQTT_Pass));
    mqtt.connect(MQTT_Name);
  } // ende MQTT Verbindung ja/nein
}
